@@ -32,7 +32,7 @@
 		if (base_stat(rc, &buf) != -1) {
 			pseudo_client_op(OP_CREAT, 0, -1, -1, tmp_template, &buf);
 		} else {
-			pseudo_debug(1, "mkdtemp (path %s) succeeded, but fstat failed (%s).\n",
+			pseudo_debug(PDBGF_CONSISTENCY, "mkdtemp (path %s) succeeded, but fstat failed (%s).\n",
 				rc, strerror(errno));
 		}
 		errno = save_errno;
