@@ -520,7 +520,7 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 		break;
 	}
 
-	/* Process rename path seperation, there are two paths old / new
+	/* Process rename path separation, there are two paths old / new
 	 * stuff into a rename, break them apart (null seperated)
 	 */
 
@@ -571,6 +571,9 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 			found_ino = 1;
 			/* note:  we have to avoid freeing this later */
 			path_by_ino = msg->path;
+			if (msg->op == OP_LINK) {
+				pseudo_debug(PDBGF_FILE, "[matches existing link]");
+			}
 		}
 	}
 
@@ -589,9 +592,8 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 		}
 		/* search on original inode -- in case of mismatch */
 		if (msg->dev && msg->ino) {
-			if (!pdb_find_file_dev(&by_ino, &row)) {
+			if (!pdb_find_file_dev(&by_ino, &row, &path_by_ino)) {
 				found_ino = 1;
-				path_by_ino = pdb_get_file_path(&by_ino);
 			}
 		}
 	}
@@ -602,7 +604,11 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 		(unsigned long long) msg_header.ino,
 		found_ino ? "+" : "-");
 
-	if (found_path) {
+	/* the sanity checks are inappropriate for DID_UNLINK, since it's
+	 * completely legitimate to have a new database entry for the
+	 * same inode.
+	 */
+	if (found_path && msg->op != OP_DID_UNLINK) {
 		/* This is a bad sign.  We should never have a different entry
 		 * for the inode...  But an inode of 0 from an EXEC is normal,
 		 * we don't track those.
@@ -657,7 +663,11 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 		}
 	}
 
-	if (found_ino) {
+	/* for OP_DID_UNLINK, the reason this op exists is that the same
+	 * inode might have been reclaimed. Don't sanity-check it, and
+	 * especially don't delete the database contents!
+	 */
+	if (found_ino && msg->op != OP_DID_UNLINK) {
 		/* Not always an absolute failure case.
 		 * If a file descriptor shows up unexpectedly and gets
 		 * fchown()d, you could have an entry giving the inode and
@@ -717,7 +727,7 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 						(unsigned long long) msg_header.ino,
 						path_by_ino ? path_by_ino : "no path",
 						msg->path);
-					}
+				}
 			}
 		} else {
 			/* I don't think I've ever seen this one. */
@@ -812,8 +822,11 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 			/* just in case find_file_path screwed up the msg */
 			msg->mode = msg_header.mode;
 		}
-		/* if the path is not known, link it */
-		if (!found_path) {
+		/* if we've never seen the file at all before, link it.
+		 * If we have it in the db by inode, but not by name,
+		 * it got fixed during the sanity checks.
+		 */
+		if (!found_path && !found_ino) {
 			pseudo_debug(PDBGF_FILE, "(new) ");
 			pdb_link_file(msg, NULL);
 		}
@@ -839,8 +852,12 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 			msg->uid = msg_header.uid;
 			msg->gid = msg_header.gid;
 		}
-		/* if the path is not known, link it */
-		if (!found_path) {
+		/* if we've never seen the file at all before, link it.
+		 * If we have it in the db by inode, but not by name,
+		 * it got fixed during the sanity checks.
+		 */
+		if (!found_path && !found_ino) {
+			pseudo_debug(PDBGF_FILE, "(new) ");
 			pdb_link_file(msg, NULL);
 		}
 		break;
